/*
 * snmpd.c - rrespond to SNMP queries from management stations
 *
 * Totally rewritten by G.Vanem <giva@bgnett.no> to
 * support DOS + Waterloo tcp/ip + Waterloo SNMP
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <io.h>
#include <process.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <snmp.h>

#include "groupvar.h"

int log_addresses = 0;
int verbose       = 0;

extern char version_descr[256];
extern char trap_sink[256];
extern char trap_community[256];

extern int init_agent_auth ();
extern int read_config ();
extern int snmp_agent_parse ();

static void read_main_config_file (void);
static int  receive    (int sd);
static int  agent_read (int sd);
static int  debug = 0;

/*
 * send a trap via snmptrap(1). ignore silently, if not avail.
 */
static void send_trap (char *host, char *comm, char *dev, int trap)
{
  char *cmd = "snmptrap";
  char trapt[20];
  int  fd;

  sprintf (trapt, "%d", trap);

  fd = open ("nul", O_RDWR);
  if (fd >= 0)
  {
    dup2 (0, fd);
    dup2 (1, fd);
    dup2 (2, fd);
  }
  execlp (cmd,cmd,host,comm,trapt,"0",dev,NULL);
}

void send_easy_trap (int trap)
{
  /* from agent/snmp_vars.c: */
  extern int snmp_enableauthentraps;	/* default: 2 == disabled */

  if (snmp_enableauthentraps != 2)
    send_trap (trap_sink, trap_community, version_descr, trap);
}

/*
 * snmp port to use:
 */
static unsigned short snmp_port = SNMP_PORT;

/*
 * path to agentinfo and configfile: 
 */
char *snmp_agentinfo  = "/etc/snmpd.agentinfo";
char *snmp_configfile = "/etc/snmpd.conf";

static void usage (void)
{
  printf ("\nuse: snmpd [-v] [-d] [-p port] [-a agentinfo] [-c configfile\n");
  exit (1);
}


int main (int argc, char **argv)
{
  int    ch, sd;
  struct sockaddr_in me;

  /*
   * usage: snmpd
   */
  while ((ch = getopt(argc,argv, "?da:c:p:v")) != EOF)
      switch (ch)
      {
        case 'd':
             snmp_dump_packet (1);
             dbug_init();
             debug++;
             break;

        case 'a':
             snmp_agentinfo = optarg;
             break;

        case 'c':
             snmp_configfile = optarg;
             break;

        case 'p':
             snmp_port = atoi (optarg);
             break;

        case 'v':
             verbose++;
             break;

        default:
             usage();
             break;
      }

  init_snmp();

  /* Set up connections
   */
  sd = socket (AF_INET, SOCK_DGRAM, 0);
  if (sd < 0)
  {
    perror ("socket");
    return (0);
  }
  me.sin_family      = AF_INET;
  me.sin_addr.s_addr = INADDR_ANY;
  me.sin_port        = htons (snmp_port);
  if (bind (sd, (struct sockaddr*)&me, sizeof(me)) != 0)
  {
    perror ("bind");
    return (0);
  }

  /* read configs, before it's too late...
   */
  read_main_config_file();

  if (init_agent_auth() < 0)
     return (1);

  if (read_config() < 0)
     return (2);

  /* set private community string, if read from config file:
   */
  {
    extern char *communities[];         /* agent/snmp_agent.c */
    extern char *public_community;	/* snmplib/snmp.c */
    extern char *private_community;	/* snmplib/snmp.c */

    if (public_community)
       communities[0] = public_community;
    if (private_community)
       communities[1] = private_community;
  }

  /* send coldstart trap via snmptrap(1) if possible
   */
  send_easy_trap (0);

  /* process incoming requests:
   */
  receive (sd);

  return (0);
}

static int receive (int sd)
{
  int    count, numfds;
  fd_set fdset;

  if (verbose)
     printf ("listening on port %d\n", snmp_port);

  while (1)
  {
    numfds = 0;
    FD_ZERO (&fdset);
    numfds = sd + 1;
    FD_SET (sd, &fdset);
    count = select_s (numfds, &fdset,NULL,NULL,NULL);
    if (count > 0)
    {
      if (FD_ISSET(sd,&fdset))
         agent_read (sd);
    }
    else
      switch (count)
      {
        case 0:
             break;
        case -1:
             if (errno == EINTR)
                continue;
             perror ("select");
             return -1;
        default:
             printf ("select returned %d\n", count);
             return -1;
      }
  }
}


static int agent_read (int sd)
{
  struct sockaddr_in from;
  int    length, out_length, fromlength;
  u_char packet[SNMP_MAX_LEN], outpacket[SNMP_MAX_LEN];

  fromlength = sizeof from;
  length = recvfrom (sd, packet, SNMP_MAX_LEN, 0,
                     (struct sockaddr*)&from, &fromlength);
  if (length == -1)
       perror ("recvfrom");
  else snmp_inpkts++;

  if (debug)
  {
    int count;

    printf ("received %d bytes from %s:\n", length, inet_ntoa(from.sin_addr));
    for (count = 0; count < length; count++)
    {
      printf ("%02X ", packet[count]);
      if ((count % 16) == 15)
         printf ("\n");
    }
    printf ("\n\n");
  }
  out_length = SNMP_MAX_LEN;
  if (snmp_agent_parse(packet,length,outpacket,&out_length,from.sin_addr))
  {
    if (debug)
    {
      int count;

      printf ("sent %d bytes to %s:\n", out_length, inet_ntoa(from.sin_addr));
      for (count = 0; count < out_length; count++)
      {
        printf ("%02X ", outpacket[count]);
        if ((count % 16) == 15)
          printf ("\n");
      }
      printf ("\n\n");
    }
    if (sendto(sd, (char*)outpacket, out_length, 0,
               (struct sockaddr*)&from, sizeof(from)) < 0)
    {
      perror ("sendto");
      return (0);
    }
    else
      snmp_outpkts++;
  }
  return (0);
}
  

/*
 * read the main-config file (if avail) and initialize the interface 
 * parameter, contact, ...
 */

static int main_config_read = 0;	/* still read ? */


/* exported to ../agent/snmp_vars.c:
 */
char  sysContact[256]   = "Unknown";
char  sysLocation[256]  = "Unknown";
char  sysName[256]      = "";
char *public_community  = 0;
char *private_community = 0;

/* port read from the configfile:
 */
int conf_snmp_port = -1;

/* trapsink host and community; setable by configfile:
 */
char trap_sink[256]      = "localhost";
char trap_community[256] = "public";
int  conf_authentraps = 0;

static void read_main_config_file (void)
{
  FILE *in;
  char *pfx;
  char  line[1024];
  
  if (main_config_read)
     return;

  main_config_read = 1;

  pfx = getenv ("SNMPCONFIGFILE");
  if (pfx && (pfx = strdup(pfx)) != NULL)
     snmp_configfile = pfx;

  if (verbose)
     printf ("reading main config file: %s\n", snmp_configfile);

  in = fopen (snmp_configfile,"r");
  if (!in)
  {
    fprintf (stderr, "warning: cannot open %s - using default paths.\n",
             snmp_configfile);
    return;
  }

  while (fgets(line, sizeof(line)-1, in))
  {
    char *key, *val;

    if (!*line || *line == '\n' || *line == '#')
       continue;

    if (line[strlen(line)-1] == '\n')
        line[strlen(line)-1] = 0;

    val = strchr (line, ':');
    if (!val)
    {
      if (strncmp(line,"view",4) &&
          strncmp(line,"user",4) &&
          strncmp (line,"community", 9))
        fprintf (stderr,
                 "warning: reading config: don't know what to do "
                 " with this line:\n\t%s\n", line);
      continue;
    }

    key = line;

    for (*val++ = 0; *val == ' ' || *val == '\t'; val++)
       continue;

    /* okey dokey; now we have a key and a value:
     */
    /* printf ("got key `%s' and val `%s'\n", key, val); */

    if (!strcmp(key, "interface"))
       printf ("ignoring keyword `interface'\n");

    else if (!strcmp(key, "system contact"))
    {
      int len = sizeof (sysContact);
      if (strlen (val) < len)
         strcpy (sysContact, val);
      else
      {
        strncpy (sysContact, val, len - 1);
        sysContact[len - 1] = 0;
      }
      if (verbose)
        printf ("added from config: system contact is %s\n", sysContact);
    }
    else if (!strcmp (key, "system location"))
    {
      int len = sizeof (sysLocation);
      if (strlen (val) < len)
        strcpy (sysLocation, val);
      else
      {
        strncpy (sysLocation, val, len - 1);
        sysLocation[len - 1] = 0;
      }
      if (verbose)
        printf ("added from config: system location is %s\n", sysLocation);
    }
    else if (!strcmp (key, "system name"))
    {
      int len = sizeof (sysName);
      if (strlen (val) < len)
         strcpy (sysName, val);
      else
      {
        strncpy (sysName, val, len - 1);
        sysName[len - 1] = 0;
      }
      if (verbose)
         printf ("added from config: system name is %s\n", sysName);
    }
    else if (!strcmp (key, "public") || !strcmp (key, "private"))
    {
      static char comm[256];
      int    err = 0;
      char  *s   = strdup (val);

      if (!s)
      {
        fprintf (stderr, "snmpd: out of mem - over and out.\n");
        exit (1);
      }
      while (isspace (comm[strlen (s) - 1]))
            s[strlen (s) - 1] = 0;
      while (*s && isspace (*s))
            s++;
      if (!*s)
      {
        fprintf (stderr,"snmpd: empty %s community string.\n", key);
        err = 1;
      }
      if (!err)
      {
        if (!strcmp(key, "private"))
             private_community = s;
        else public_community = s;
      }
      else
        fprintf (stderr, "snmpd: default %s community will be used.\n", key);
    }
    else if (!strcmp(key,"port"))
    {
      unsigned short p = atoi (val);
      if (!val)
        fprintf (stderr, "snmpd: no valid port found.\n");
      else
      {
        conf_snmp_port = p;
        if (verbose)
           printf ("port set to %d\n", conf_snmp_port);
      }
    }
    else if (!strcmp (key, "trap sink"))
    {
      int len = sizeof (trap_sink);
      if (strlen (val) < len)
        strcpy (trap_sink, val);
      else
      {
        strncpy (trap_sink, val, len - 1);
        trap_sink[len - 1] = 0;
      }
      if (verbose)
         printf ("added from config: trap sink addess is %s\n",trap_sink);
    }
    else if (!strcmp (key, "trap community"))
    {
      int len = sizeof (trap_community);
      if (strlen (val) < len)
         strcpy (trap_community, val);
      else
      {
        strncpy (trap_community, val, len - 1);
        trap_community[len - 1] = 0;
      }
      if (verbose)
         printf ("added from config: trap community string is %s\n",trap_community);
    }
    else if (!strcmp (key, "authentraps"))
    {
      if (!strcmp (val, "yes"))
           conf_authentraps = 1;
      else if (!strcmp (val, "no"))
           conf_authentraps = 2;
      else fprintf (stderr, "warning: reading config: unknown val for %s\n", key);
      if (verbose)
        printf ("added from config: authentraps set to %s\n", val);
    }
    else
      fprintf (stderr, "warning: reading config: unknown key `%s'\n", key);
  }
  fclose (in);
}
