/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Changes by G. Vanem 1998
 * Adapted to Waterloo tcp/ip
 **********************************************************************/

#include "snmp_lib.h"

#ifdef __TURBOC__
#pragma warn -pro
#endif


/* 
 * RFC 1906: Transport Mappings for SNMPv2
 */

oid default_enterprise[] = { 1, 3, 6, 1, 4, 1, 3, 1, 1 };
                          /* enterprises.cmu.systems.cmuSNMP */

#define DEFAULT_COMMUNITY   "public"
#define DEFAULT_RETRIES     4
#define DEFAULT_TIMEOUT     2000000L
#define DEFAULT_REMPORT     SNMP_PORT
#define DEFAULT_ENTERPRISE  default_enterprise
#define DEFAULT_TIME        0

struct session_list *Sessions = NULL;

/* 
 * Get initial request ID for all transactions.
 */
static int Reqid = 0;
static int default_s_port = 0;

static void *snmp_sess_pointer (struct snmp_session *session);

static void init_snmp (void)
{
  struct servent *sp;

  if (Reqid)
     return;

  Reqid = 1; 
  if ((sp = getservbyname ("snmp", "udp")) != NULL)
     default_s_port = sp->s_port;

#ifdef __BORLANDC__
  srand (time(NULL));
  Reqid = random (INT_MAX);
#else
  srandom (time(NULL));
  Reqid = random();
#endif
}


/* 
 * Free each element in the input request list.
 */
static void free_request_list (struct request_list *rp)
{
  struct request_list *orp;

  while (rp)
  {
    orp = rp;
    rp  = rp->next_request;
    if (orp->pdu)
       snmp_free_pdu (orp->pdu);
    free (orp);
  }
}

/*
 * Sets up the session with the snmp_session information provided
 * by the user.  Then opens and binds the necessary UDP port.
 * A handle to the created session is returned (this is different than
 * the pointer passed to snmp_open()).  On any error, NULL is returned
 * and snmp_errno is set to the appropriate error code.
 */
struct snmp_session *snmp_open (struct snmp_session *session)
{
  struct session_list *slp;

  slp = (struct session_list*) snmp_sess_open (session);
  if (!slp)
     return (NULL);

  slp->next = Sessions;
  Sessions  = slp;
  return (slp->session);
}

/*
 * The "spin-free" version of snmp_open
 */
void *snmp_sess_open (struct snmp_session *in_session)
{
  struct session_list          *slp;
  struct snmp_session          *session;
  struct snmp_internal_session *isp;
  static DWORD fd = 0;

  int         tmplen;
  BYTE       *cp, *tmpcp;
  DWORD       addr;
  WORD        port;
  udp_Socket *sock;

  if (Reqid == 0)
     init_snmp();

  /* Copy session structure and link into list
   */
  slp = calloc (sizeof(*slp),1);
  if (!slp)
  {
    snmp_set_api_error (SNMPERR_OS_ERR);
    in_session->s_snmp_errno = SNMPERR_OS_ERR;
    return (NULL);
  }

  /* Internal session
   */
  isp = calloc (sizeof(*isp),1);
  if (!isp)
  {
    snmp_set_api_error (SNMPERR_OS_ERR);
    in_session->s_snmp_errno = SNMPERR_OS_ERR;
    snmp_sess_close (slp);
    return (NULL);
  }

  /* Set up udp socket
   */
  sock = malloc (sizeof(*sock));  
  if (!sock)
  {
    snmp_set_api_error (SNMPERR_OS_ERR);
    in_session->s_snmp_errno = SNMPERR_OS_ERR;
    snmp_sess_close (slp);
    return (NULL);
  }

  isp->socket       = (void*)sock;
  slp->internal     = isp;
  slp->internal->sd = ++fd;    /* dummy fd for WatTCP */
  if (fd > FD_SETSIZE)         /* range 1..FD_SETSIZE */
      fd = 1;

  /* The actual session
   */
  slp->session = malloc (sizeof (struct snmp_session));
  if (!slp->session)
  {
    snmp_set_api_error (SNMPERR_OS_ERR);
    in_session->s_snmp_errno = SNMPERR_OS_ERR;
    snmp_sess_close (slp);
    return (NULL);
  }

  memcpy (slp->session, in_session, sizeof(struct snmp_session));
  session = slp->session;
  slp->session->socket = (void*)sock;

  /*
   * session now points to the new structure that still contains pointers to
   * data allocated elsewhere.  Some of this data is copied to space malloc'd
   * here, and the pointer replaced with the new one.
   */
  if (session->peername)
  {
    cp = strdup (session->peername);
    if (!cp)
    {
      snmp_set_api_error (SNMPERR_OS_ERR);
      in_session->s_snmp_errno = SNMPERR_OS_ERR;
      snmp_sess_close (slp);
      return (NULL);
    }
    session->peername = (char*) cp;
  }

  /* Fill in defaults if necessary
   */
  tmpcp  = session->community;
  tmplen = session->community_len;
  if (tmplen == 0)
  {
    tmpcp  = DEFAULT_COMMUNITY;
    tmplen = strlen (tmpcp);
  }
  cp = malloc (tmplen+1);
  if (!cp)
  {
    snmp_set_api_error (SNMPERR_OS_ERR);
    in_session->s_snmp_errno = SNMPERR_OS_ERR;
    snmp_sess_close (slp);
    return (NULL);
  }

  memmove (cp, tmpcp, tmplen);
  session->community_len = tmplen;
  session->community     = cp;  /* replace ptr with ptr to new data */

  if (session->retries == SNMP_DEFAULT_RETRIES ||
      session->retries == OLD_SNMP_DEFAULT_RETRIES)
     session->retries = DEFAULT_RETRIES;

  if (session->timeout == SNMP_DEFAULT_TIMEOUT ||
      session->timeout == OLD_SNMP_DEFAULT_TIMEOUT)
     session->timeout = DEFAULT_TIMEOUT;

  isp->requests = NULL;

  if (session->peername != SNMP_DEFAULT_PEERNAME)
  {
    if ((addr = inet_addr(session->peername)) != INADDR_NONE)
       isp->addr.sin_addr.s_addr = addr;
    else
    {
      addr = resolve (session->peername);
      if (!addr)
      {
        snmp_errno = SNMPERR_BAD_ADDRESS;
        in_session->s_snmp_errno = SNMPERR_BAD_ADDRESS;
     // in_session->s_errno      = h_errno;
        snmp_sess_close (slp);
        return (NULL);
      }
      isp->addr.sin_addr.s_addr = htonl (addr);
    }
    isp->addr.sin_family = AF_INET;

    if (session->remote_port == SNMP_DEFAULT_REMPORT)
    {
      if (default_s_port)
           isp->addr.sin_port = default_s_port;
      else isp->addr.sin_port = htons (SNMP_PORT);
    }
    else
      isp->addr.sin_port = htons (session->remote_port);
  }
  else
    isp->addr.sin_addr.s_addr = SNMP_DEFAULT_ADDRESS;

  addr = ntohl (isp->addr.sin_addr.s_addr);
  port = ntohs (isp->addr.sin_port);
  if (!udp_open(isp->socket, 0, addr, port, NULL))
  {
    snmp_errno = SNMPERR_BAD_LOCPORT;
    in_session->s_snmp_errno = SNMPERR_BAD_LOCPORT;
    in_session->s_errno      = ENETUNREACH;
    snmp_sess_close (slp);
    return (NULL);
  }
  return (slp);
}


static void _snmp_free (void *cp)
{
  if (cp)
    free (cp);
}

/* 
 * Close the input session.  Frees all data allocated for the session,
 * dequeues any pending requests, and closes any sockets allocated for
 * the session.  Returns 0 on error, 1 otherwise.
 */       
int snmp_sess_close (void *sessp)
{
  struct session_list *slp = (struct session_list *) sessp;

  if (!slp)
     return (0);

  if (slp->internal)
  {
    if (slp->internal->socket)
       sock_close (slp->internal->socket);
    free_request_list (slp->internal->requests);
  }

  _snmp_free (slp->session->peername);
  _snmp_free (slp->session->community);
  _snmp_free (slp->internal->socket);
  _snmp_free (slp->session);
  _snmp_free (slp->internal);
  _snmp_free (slp);
  return (1);
}

int snmp_close (struct snmp_session *session)
{
  struct session_list *slp  = NULL;
  struct session_list *oslp = NULL;

  if (Sessions->session == session)
  {                                /* If first entry */
    slp = Sessions;
    Sessions = slp->next;
  }
  else
  {
    for (slp = Sessions; slp; slp = slp->next)
    {
      if (slp->session == session)
      {
        if (oslp)                   /* if we found entry that points here */
          oslp->next = slp->next;   /* link around this entry */
        break;
      }
      oslp = slp;
    }
  }

  /* If we found the session, free all data associated with it */
  if (!slp)
  {
    snmp_errno = SNMPERR_BAD_SESSION;
    return (0);
  }
  return snmp_sess_close ((void*)slp);
}

/* 
 * Takes a session and a pdu and serializes the ASN PDU into the area
 * pointed to by packet.  out_length is the size of the data area available.
 * Returns the length of the encoded packet in out_length.  If an error
 * occurs, -1 is returned.  If all goes well, 0 is returned.
 */
int snmp_build (struct snmp_session *session, struct snmp_pdu *pdu,
                BYTE *packet, int *out_length)
{
  BYTE *bufp = snmp_msg_Encode (packet, out_length,
                                session->community, session->community_len,
                                session->Version, pdu);

  API_DEBUG2 ("snmp_build(): Packet len difference %d, returning %d\n",
              (int)(bufp-packet), *out_length);

  if (!bufp)
     return (-1);
  return (0);
}

/* 
 * Parses the packet received on the input session, and places the data into
 * the input pdu.  length is the length of the input packet.  If any errors
 * are encountered, NULL is returned.  If not, the community is.
 */
BYTE *snmp_parse (struct snmp_session *session,
                    struct snmp_pdu     *pdu,
                    BYTE *data, int length)
{
  BYTE  Community[128];
  int   CommunityLen = sizeof(Community);
  BYTE *bufp;

  /* Decode the entire message.
   */
  data = snmp_msg_Decode (data, &length, Community, &CommunityLen,
                          &session->Version, pdu);
  if (!data)
     return (NULL);

  bufp = malloc (CommunityLen+1);
  if (!bufp)
     return (NULL);

  strcpy (bufp, Community);
  return (bufp);
}

/* 
 * Sends the input pdu on the session after calling snmp_build to create
 * a serialized packet.  If necessary, set some of the pdu data from the
 * session defaults.  Add a request corresponding to this pdu to the list
 * of outstanding requests on this session, then send the pdu.
 * Returns the request id of the generated packet if applicable, otherwise 1.
 * On any error, 0 is returned.
 * The pdu is freed by snmp_send() unless a failure occured.
 */
int snmp_send (struct snmp_session *session, struct snmp_pdu *pdu)
{
  return snmp_async_send (session, pdu, NULL, NULL);
}

int snmp_sess_send (void *sessp, struct snmp_pdu *pdu)
{
  return snmp_sess_async_send (sessp, pdu, NULL, NULL);
}

/* 
 * int snmp_async_send(session, pdu, callback, cb_data)
 *     struct snmp_session *session;
 *     struct snmp_pdu  *pdu;
 *     snmp_callback callback;
 *     void   *cb_data;
 * 
 * Sends the input pdu on the session after calling snmp_build to create
 * a serialized packet.  If necessary, set some of the pdu data from the
 * session defaults.  Add a request corresponding to this pdu to the list
 * of outstanding requests on this session and store callback and data, 
 * then send the pdu.
 * Returns the request id of the generated packet if applicable, otherwise 1.
 * On any error, 0 is returned.
 * The pdu is freed by snmp_send() unless a failure occured.
 */
int snmp_async_send (struct snmp_session *session, struct snmp_pdu *pdu,
                     snmp_callback callback, void *cb_data)
{
  void *sessp = snmp_sess_pointer (session);

  if (!sessp)
     return (0);

  return snmp_sess_async_send (sessp, pdu, callback, cb_data);
}

int snmp_sess_async_send (void *sessp, struct snmp_pdu *pdu,
                          snmp_callback callback, void *cb_data)
{
  struct session_list          *slp     = (struct session_list *) sessp;
  struct snmp_internal_session *isp     = slp->internal;
  struct snmp_session          *session = slp->session;
  struct request_list          *rp;
  struct timeval                tv;
  BYTE   packet[PACKET_LENGTH];
  int    length = PACKET_LENGTH;

  session->s_snmp_errno = 0;
  session->s_errno      = 0;

  if (pdu->command == SNMP_PDU_GET      ||
      pdu->command == SNMP_PDU_GETNEXT  ||
      pdu->command == SNMP_PDU_RESPONSE ||
      pdu->command == SNMP_PDU_SET)
  {
    if (pdu->reqid == SNMP_DEFAULT_REQID)
        pdu->reqid = ++Reqid;
  }
  else if (pdu->command == SNMP_PDU_INFORM  ||
           pdu->command == SNMP_PDU_GETBULK ||
           pdu->command == SNMP_PDU_V2TRAP)
  {
    if (session->Version != SNMP_VERSION_2)
    {
      fprintf (stderr, "Cant send SNMPv2 PDU's in SNMP message.\n");
      snmp_errno = SNMPERR_GENERR;
      session->s_snmp_errno = SNMPERR_GENERR;
      return (0);
    }
    if (pdu->reqid == SNMP_DEFAULT_REQID)
        pdu->reqid = ++Reqid;
  }
  else
  {
    /* fill in trap defaults
     */
    pdu->reqid = 1;       /* give a bogus non-error reqid for traps */
    if (pdu->enterprise_length == SNMP_DEFAULT_ENTERPRISE_LENGTH)
    {
      pdu->enterprise = malloc (sizeof(DEFAULT_ENTERPRISE));
      memcpy (pdu->enterprise, DEFAULT_ENTERPRISE, sizeof(DEFAULT_ENTERPRISE));
      pdu->enterprise_length = DIM (DEFAULT_ENTERPRISE);
    }
    if (pdu->time == SNMP_DEFAULT_TIME)
        pdu->time = DEFAULT_TIME;
  }

  if (pdu->address.sin_addr.s_addr == SNMP_DEFAULT_ADDRESS)
  {
    if (isp->addr.sin_addr.s_addr != SNMP_DEFAULT_ADDRESS)
       memcpy (&pdu->address, &isp->addr, sizeof(pdu->address));
    else
    {
      (*_printf) ("No remote IP address specified\r\n");
      snmp_errno = SNMPERR_BAD_ADDRESS;
      session->s_snmp_errno = SNMPERR_BAD_ADDRESS;
      return (0);
    }
  }

  if (snmp_build (session, pdu, packet, &length) < 0)
  {
    (*_printf) ("Error building packet\r\n");
    snmp_errno = SNMPERR_GENERR;
    session->s_snmp_errno = SNMPERR_GENERR;
    return (0);
  }

  snmp_dump (packet, length, "sending", pdu->address.sin_addr);
  gettimeofday (&tv, NULL);

  sock_fastwrite (isp->socket,(const char*)packet,length);

  API_DEBUG2 ("Sent PDU %s, Reqid %d\n", snmp_pdu_type(pdu), pdu->reqid);

  if (pdu->command == SNMP_PDU_GET     ||
      pdu->command == SNMP_PDU_GETNEXT ||
      pdu->command == SNMP_PDU_SET     ||
      pdu->command == SNMP_PDU_GETBULK ||
      pdu->command == SNMP_PDU_INFORM)
  {
    API_DEBUG1 ("Setting up to receive a response for reqid %d\n", pdu->reqid);

    /* set up to expect a response
     */
    rp = calloc (sizeof(*rp),1);
    if (!rp)
    {
      snmp_errno = SNMPERR_GENERR;
      session->s_snmp_errno = SNMPERR_GENERR;
      return (0);
    }

    rp->next_request = isp->requests;
    isp->requests    = rp;

    rp->pdu          = pdu;
    rp->request_id   = pdu->reqid;
    rp->retries      = 0;
    rp->timeout      = session->timeout;
    rp->time         = tv;
    tv.tv_usec      += rp->timeout;
    tv.tv_sec       += tv.tv_usec / 1000000L;
    tv.tv_usec      %= 1000000L;
    rp->expire       = tv;
  }      
  ARGSUSED (callback);
  ARGSUSED (cb_data);
  return (pdu->reqid);
}

/* 
 * Checks to see if any of the fd's set in the fdset belong to
 * snmp.  Each socket with it's fd set has a packet read from it
 * and snmp_parse is called on the packet received.  The resulting pdu
 * is passed to the callback routine for that session.  If the callback
 * routine returns successfully, the pdu and it's request are deleted.
 */
void snmp_read (fd_set *fdset)
{
  struct session_list *slp;

  for (slp = Sessions; slp; slp = slp->next)
  {
    if (FD_ISSET (slp->internal->sd, fdset))
       snmp_sess_read (slp, fdset);
  }
}

/*
 * Same as snmp_read, but works just one session.
 */
void snmp_sess_read (void *sessp, fd_set * fdset)
{
  struct session_list          *slp = (struct session_list *) sessp;
  struct snmp_session          *sp;
  struct snmp_internal_session *isp;
  struct sockaddr_in            from;
  struct snmp_pdu              *pdu;
  struct request_list          *rp, *orp = NULL;

  int    length;
  BYTE   packet[PACKET_LENGTH];
  BYTE  *bufp;

  if (!(FD_ISSET (slp->internal->sd, fdset)))
     return;

  if (!slp->internal->socket || !sock_dataready(slp->internal->socket))
     return;

  sp  = slp->session;
  isp = slp->internal;
  sp->s_snmp_errno = 0;
  sp->s_errno      = 0;


  length = sock_fastread (isp->socket,(char*)packet,PACKET_LENGTH);
  if (length <= 0)
  {
    snmp_errno = SNMPERR_GENERR;
    sp->s_snmp_errno = SNMPERR_GENERR;
    sp->s_errno = errno;
    return;
  }

  _getpeername (isp->socket,&from,NULL);

  snmp_dump (packet, length, "received", from.sin_addr);

  pdu = snmp_pdu_create (0);
  pdu->address = from;

  /* Parse the incoming packet
   */
  bufp = snmp_parse (sp, pdu, packet, length);
  if (!bufp)
  {
    snmp_free_pdu (pdu);
    return;
  }
  if (sp->community)
     free (sp->community);

  sp->community     = bufp;
  sp->community_len = strlen (bufp);

  API_DEBUG2 ("Read PDU %s, ReqId %d\n", snmp_pdu_type(pdu), pdu->reqid);

  if (pdu->command == SNMP_PDU_RESPONSE)
  {
    for (rp = isp->requests; rp; rp = rp->next_request)
    {
      if (rp->request_id == pdu->reqid)
      {
        API_DEBUG1 ("ReqId %d:  Calling callback\n", pdu->reqid);

        if ((!sp->callback) ||
            (*sp->callback) (RECEIVED_MESSAGE, sp, pdu->reqid, pdu,
                             sp->callback_magic) == 1)
        {
          /* successful, so delete request
           */
          API_DEBUG1 ("ReqId %d: Success. Removing ReqId.\n", pdu->reqid);
          orp = rp;
          if (isp->requests == orp)
          {
            /* first in list */
            isp->requests = orp->next_request;
          }
          else
          {
            for (rp = isp->requests; rp; rp = rp->next_request)
                if (rp->next_request == orp)
                {
                  /* link around it */
                  rp->next_request = orp->next_request;
                  break;
                }
          }
          snmp_free_pdu (orp->pdu);
          free (orp);

          /* there shouldn't be another req with the same reqid
           */
          break;
        }
      }
    }
  }
  else if (pdu->command == SNMP_PDU_GET     ||
           pdu->command == SNMP_PDU_GETNEXT ||
           pdu->command == TRP_REQ_MSG      ||
           pdu->command == SNMP_PDU_SET     ||
           pdu->command == SNMP_PDU_GETBULK ||
           pdu->command == SNMP_PDU_INFORM  ||
           pdu->command == SNMP_PDU_V2TRAP)
  {
    if (sp->callback)
      (*sp->callback) (RECEIVED_MESSAGE, sp, pdu->reqid,
                       pdu, sp->callback_magic);
  }
  snmp_free_pdu (pdu);
}

/* 
 * Returns info about what snmp requires from a select statement.
 * numfds is the number of fds in the list that are significant.
 * All file descriptors opened for SNMP are OR'd into the fdset.
 * If activity occurs on any of these file descriptors, snmp_read
 * should be called with that file descriptor set
 *
 * The timeout is the latest time that SNMP can wait for a timeout.  The
 * select should be done with the minimum time between timeout and any other
 * timeouts necessary.  This should be checked upon each invocation of select.
 * If a timeout is received, snmp_timeout should be called to check if the
 * timeout was for SNMP.  (snmp_timeout is idempotent)
 *
 * Block is 1 if the select is requested to block indefinitely, rather than
 * time out.
 * If block is input as 1, the timeout value will be treated as undefined,
 * but it must be available for setting in snmp_select_info.  On return, if
 * block is true, the value of timeout will be undefined.
 *
 * snmp_select_info returns the number of open sockets.  (i.e. The number
 * of sessions open)
 */
int snmp_select_info (int *numfds, fd_set *fdset,
                      struct timeval *timeout, int *block)
{
  return snmp_sess_select_info (NULL, numfds, fdset, timeout, block);
}

/*
 * Same as snmp_select_info, but works just one session.
 */
int snmp_sess_select_info (void *sessp, int *numfds, fd_set *fdset,
                           struct timeval *timeout, int *block)

{
  struct session_list          *slptest = (struct session_list*) sessp;
  struct session_list          *slp;
  struct snmp_internal_session *isp;
  struct request_list          *rp;
  struct timeval                now, earliest;
  int    active   = 0;
  int    requests = 0;

  timerclear (&earliest);
  /*
   * For each request outstanding, add it's socket to the fdset,
   * and if it is the earliest timeout to expire, mark it as lowest.
   * If a single session is specified, do just for that session.
   */
  if (slptest)
       slp = slptest;
  else slp = Sessions;

  for ( ; slp; slp = slp->next)
  {     
    active++;
    isp = slp->internal;
    if ((isp->sd + 1) > *numfds)
       *numfds = (isp->sd + 1);
    FD_SET (isp->sd, fdset);

    API_DEBUG1 ("Adding port %d\n", isp->sd);
    if (isp->requests)
    {
      /* found another session with outstanding requests
       */
      requests++;
      for (rp = isp->requests; rp; rp = rp->next_request)
      {
        if (!timerisset (&earliest) ||
            timercmp (&rp->expire, &earliest, <))
          earliest = rp->expire;
      }
    }
    if (slp == slptest)
       break;
  }

  API_DEBUG2 ("Select Info:  %d active, %d requests pending.\n",
              active, requests);

  if (requests == 0)                /* if none are active, skip arithmetic */
  {
    *block = 1;   /* can block - timeout value is undefined if no requests */
    return (active);
  }

  /*
   * Now find out how much time until the earliest timeout.  This
   * transforms earliest from an absolute time into a delta time, the
   * time left until the select should timeout.
   */
  gettimeofday (&now,NULL);

  API_DEBUG1 ("Select Info: Earliest request expires in %d secs.\n",
              earliest.tv_sec);

  earliest.tv_sec--;                /* adjust time to make arithmetic easier */
  earliest.tv_usec += 1000000L;
  earliest.tv_sec  -= now.tv_sec;
  earliest.tv_usec -= now.tv_usec;

  while (earliest.tv_usec >= 1000000L)
  {
    earliest.tv_usec -= 1000000L;
    earliest.tv_sec  += 1;
  }
  if (earliest.tv_sec < 0)
  {
    earliest.tv_sec  = 0;
    earliest.tv_usec = 0;
  }

  /* if it was blocking before or our delta time is less, reset timeout
   */
  if (*block == 1 || timercmp (&earliest, timeout, <))
  {
    *timeout = earliest;
    *block   = 0;
  }
  return (active);
}

/* 
 * snmp_timeout should be called whenever the timeout from snmp_select_info 
 * expires, but it is idempotent, so snmp_timeout can be polled (probably a 
 * cpu expensive proposition).  snmp_timeout checks to see if any of the 
 * sessions have an outstanding request that has timed out.  If it finds one 
 * (or more), and that pdu has more retries available, a new packet is formed
 * from the pdu and is resent.  If there are no more retries available, the 
 * callback for the session is used to alert the user of the timeout.
 */
void snmp_timeout (void)
{
  struct session_list *slp;

  for (slp = Sessions; slp; slp = slp->next)
     snmp_sess_timeout (slp);
}

void snmp_sess_timeout (void *sessp)
{
  struct session_list          *slp = (struct session_list *) sessp;
  struct snmp_session          *sp  = slp->session;
  struct snmp_internal_session *isp = slp->internal;
  struct request_list          *rp, *orp = NULL;
  struct request_list          *freeme   = NULL;
  struct timeval now;

  gettimeofday (&now,NULL);

  API_DEBUG1 ("Checking session %s\n", sp->peername ? sp->peername : "<NULL>");

  for (rp = isp->requests; rp; rp = rp->next_request)
  {
    API_DEBUG4 ("Checking session request %d, expire at %u, Retry %d/%d\n",
                rp->request_id, rp->expire.tv_sec, rp->retries, sp->retries);
    if (freeme)
    {
      free (freeme);
      freeme = NULL;
    }

    if (timercmp (&rp->expire, &now, <))
    {
      API_DEBUG0 ("Expired.\n");

      /* this timer has expired
       */
      if (rp->retries >= sp->retries)
      {
        /* No more chances, delete this entry
         */
        if (sp->callback)
          (*sp->callback) (TIMED_OUT, sp, rp->pdu->reqid,
                           rp->pdu, sp->callback_magic);
        if (orp)
             orp->next_request = rp->next_request;
        else isp->requests     = rp->next_request;

        snmp_free_pdu (rp->pdu);
        freeme = rp;
        continue;             /* don't update orp below */
      }
      else
      {
        BYTE   packet[PACKET_LENGTH];
        int    length = PACKET_LENGTH;
        struct timeval tv;
    
        rp->retries++;

        API_DEBUG2 ("Retransmitting (%d, timeout %ld).\n",
                    rp->retries, rp->timeout);

        if (snmp_build (sp, rp->pdu, packet, &length) < 0)
        {
          /* Error for application to process */
          break;
        }
        snmp_dump (packet, length, "sending", rp->pdu->address.sin_addr);
        gettimeofday (&tv,NULL);

        sock_fastwrite (isp->socket,(const char*)packet,length);

        rp->time    = tv;
        tv.tv_usec += rp->timeout;
        tv.tv_sec  += tv.tv_usec / 1000000L;
        tv.tv_usec %= 1000000L;
        rp->expire  = tv;
      }
    }
    orp = rp;
  }

  if (freeme)
  {
    free (freeme);
    freeme = NULL;
  }
}


/*
 * Print some API stats
 */
void snmp_api_stats (void *outP)
{
  struct session_list          *slp;
  struct request_list          *rp;
  struct snmp_internal_session *isp;

  FILE  *out      = (FILE*) outP;
  int    active   = 0;
  int    requests = 0;
  int    count    = 0;
  int    rcount   = 0;

  fprintf (out, "LIBSNMP: Session List Dump\n");
  fprintf (out, "LIBSNMP: ----------------------------------------\n");

  for (slp = Sessions; slp; slp = slp->next)
  {
    isp = slp->internal;
    active++;
    count++;
    fprintf (out, "LIBSNMP: %2d: Host %s\n", count,
             slp->session->peername ? slp->session->peername : "NULL");

    if (isp->requests)
    {
      /* found another session with outstanding requests
       */
      requests++;
      rcount = 0;
      for (rp = isp->requests; rp; rp = rp->next_request)
      {
        rcount++;
        {
          struct hostent *hp = gethostbyaddr ((const char*)&rp->pdu->address,
                                              sizeof(DWORD), AF_INET);
          fprintf (out, "LIBSNMP: %2d: ReqId %d (%s) (%s)\n",
                   rcount, rp->request_id, snmp_pdu_type (rp->pdu),
                   hp ? hp->h_name : "NULL");
        }
      }
    }
    fprintf (out, "LIBSNMP: ----------------------------------------\n");
  }
  fprintf (out, "LIBSNMP: Session List: %d active, %d have requests pending.\n",
           active, requests);
}

/*
 * snmp_sess_error - same as snmp_error for single session API use.
 */
void snmp_sess_error (void *sessp, int *p_errno,
                      int *p_snmp_errno, char **p_str)
{
  struct session_list *slp = (struct session_list*) sessp;

  if (slp && slp->session)
     snmp_error (slp->session, p_errno, p_snmp_errno, p_str);
}


void snmp_sess_init (struct snmp_session *session)
{
  init_snmp();

  /* initialize session to default values
   */
  memset (session, 0, sizeof(*session));

  session->remote_port = SNMP_DEFAULT_REMPORT;
  session->timeout     = SNMP_DEFAULT_TIMEOUT;
  session->retries     = SNMP_DEFAULT_RETRIES;
}

/* 
 * returns NULL or internal pointer to session
 * use this pointer for the other snmp_sess* routines,
 * which guarantee action will occur ONLY for this given session.
 */
void *snmp_sess_pointer (struct snmp_session *session)
{
  struct session_list *slp;

  for (slp = Sessions; slp; slp = slp->next)
      if (slp->session == session)
         break;

  if (!slp)
  {
    snmp_errno = SNMPERR_BAD_SESSION;
    return (NULL);
  }
  return ((void*)slp);
}

/* 
 * Input : an opaque pointer, returned by snmp_sess_open.
 * returns NULL or pointer to session.
 */
struct snmp_session *snmp_sess_session (void *sessp)
{
  struct session_list *slp = (struct session_list*) sessp;

  if (!slp)
     return (NULL);
  return (slp->session);
}
