/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Changes by G. Vanem 1998
 * Adapted to Waterloo tcp/ip
 **********************************************************************/

#include "snmp_lib.h"

/*=============================================================================
 *
 *      Name            : snmp_get_internal_session
 *
 *        Description        : Get the internal_session of the session
 *
 *        Parameters        : 
 *                - struct snmp_session *session           => session
 *                + struct snmp_internal_session *internal => internal session
 *
 *        Return value: 
 *                0        => I don't find it
 *                1        => I find it
 *
 *=============================================================================
 */
int snmp_get_internal_session (struct snmp_session *session,
                               struct snmp_internal_session *internal)
{
  struct session_list          *slp;
  struct snmp_internal_session *isp = NULL;

  for (slp = Sessions; slp; slp = slp->next)
  {
    if (slp->session == session)
    {
      isp = slp->internal;
      break;
    }
  }

  if (!isp)
  {
    snmp_errno = SNMPERR_BAD_SESSION;
    return (0);
  }
  *internal = *isp;
  return (1);
}

/*=============================================================================
 *
 *      Name            : snmp_get_request_list
 *
 *        Description        : Get the Request List of the session
 *
 *        Parameters        : 
 *                - struct snmp_session *session        => session
 *              + struct request_list *list     => request list
 *
 *        Return value: 
 *                0        => I don't find it
 *                1        => I find it
 *
 *=============================================================================
 */
int snmp_get_request_list (struct snmp_session *session,
                           struct request_list *list)
{
  struct session_list *slp;
  struct request_list *rqst = NULL;

  for (slp = Sessions; slp; slp = slp->next)
  {
    if (slp->session == session)
    {
      rqst = slp->internal->requests;
      break;
    }
  }

  if (!rqst)
  {
    snmp_errno = SNMPERR_BAD_SESSION;
    return (0);
  }
  *list = *rqst;
  return (1);
}


/*=============================================================================
 *
 *      Name            : snmp_modif_param_session
 *
 *        Description        : Modification of the parameters of the session execpt
 *                        peername
 *                        remote_port
 *                       local_port
 *
 *        Parameters        : 
 *                - struct snmp_session *session        => session
 *
 *        Return value: 
 *                0        => I don't find it
 *                1        => I find it
 *
 *=============================================================================
 */
int snmp_modif_param_session (struct snmp_session *session)
{
  struct session_list *slp;
  char  *cp;

  for (slp = Sessions; slp; slp = slp->next)
      if (slp->session == session)
         break;

  if (!slp)
  {
    snmp_errno = SNMPERR_BAD_SESSION;
    return (0);
  }

  if (strcmp(slp->session->community,session->community))
  {
    cp = malloc (session->community_len + 1);
    if (cp)
    {
      strcpy (cp,session->community);
      free (slp->session->community);
      slp->session->community_len = session->community_len;
      slp->session->community     = (BYTE *) cp;
    }
    else
    {
      snmp_errno = SNMPERR_OS_ERR;
      return (0);
    }
  }
  if (slp->session->Version != session->Version)
      slp->session->Version = session->Version;

  if (slp->session->retries != session->retries)
      slp->session->retries = session->retries;

  if (slp->session->timeout != session->timeout)
      slp->session->timeout = session->timeout;

  return (1);
}


/*=============================================================================
 *
 *      Name            : snmp_get_socket_session
 *
 *        Description        : retrune the socket number of the session_
 *
 *        Parameters        : 
 *                - struct snmp_session *session        => session_
 *
 *        Return value: 
 *                >0        => Socket number
 *                -1        => I find it
 *
 *=============================================================================
 */
int snmp_get_socket_session (struct snmp_session *session_)
{
  struct session_list *slp;

  for (slp = Sessions; slp; slp = slp->next)
      if (slp->session == session_)
         break;

  if (!slp)
  {
    snmp_errno = SNMPERR_BAD_SESSION;
    return (-1);
  }
  return (slp->internal->sd);
}



/*=============================================================================
 *
 * Name        : snmp_select_info_session
 *
 * Description : return the next time of the closest request of the session
 *               if it exists...
 *
 *        Parameters  : 
 *    - struct snmp_session* session   => session
 *    - struct timeval*      timeout   => time of the next TimeOut Request
 *
 * Return value: 
 *    -1 => We don't find the session.
 *     0 => No more request for this sesion.
 *    >0 => Number of request for this session.
 *
 *=============================================================================
 */
int snmp_select_info_session (struct snmp_session *session_, struct timeval *timeout)
{
  struct session_list          *slp;
  struct snmp_internal_session *isp;
  struct request_list          *rp;
  struct timeval                now, earliest;
  int    requests = 0;

  for (slp = Sessions; slp; slp = slp->next)
      if (slp->session == session_)
         break;

  if (!slp)
  {
    snmp_errno = SNMPERR_BAD_SESSION;
    return (-1);
  }

  timerclear (&earliest);
  isp = slp->internal;

  if (isp->requests)
  {
    requests++;
    /* found another session with outstanding requests */
    for (rp = isp->requests; rp; rp = rp->next_request)
    {
      if (!timerisset (&earliest) || timercmp (&rp->expire, &earliest, <))
        earliest = rp->expire;
    }
  }
  API_DEBUG1 ("Select Info:  %d requests pending.\n", requests);

  if (requests == 0)          /* if none are active, skip arithmetic */
     return (0);

  gettimeofday (&now,NULL);

  earliest.tv_sec--;          /* adjust time to make arithmetic easier */
  earliest.tv_usec += 1000000L;
  earliest.tv_sec  -= now.tv_sec;
  earliest.tv_usec -= now.tv_usec;

  while (earliest.tv_usec >= 1000000L)
  {
    earliest.tv_usec -= 1000000L;
    earliest.tv_sec  += 1;
  }
  if (earliest.tv_sec < 0)
  {
    earliest.tv_sec  = 0;
    earliest.tv_usec = 0;
  }
  *timeout = earliest;

  API_DEBUG1 ("Select Info: Earliest request expires in %d secs.\n",
              earliest.tv_sec);
  return (requests);
}


/*=============================================================================
 *
 * Name        : snmp_timeout_session
 *
 * Description : retranmit the TimeOut request of the session_ if it is possible,
 *               unless call the callBack and delete it from the API.
 *
 *        Parameters  : 
 *    - struct snmp_session *session        => session
 *
 * Return value: 
 *    0 => We don't find the session.
 *    1 => We find the session.
 *
 *=============================================================================
 */
int snmp_timeout_session (struct snmp_session *session_)
{
  struct session_list          *slp;
  struct snmp_session          *sp;
  struct snmp_internal_session *isp;
  struct request_list          *rp, *orp;
  struct request_list          *freeme = NULL;
  struct timeval               now;

  for (slp = Sessions; slp; slp = slp->next)
      if (slp->session == session_)
         break;

  if (!slp)
  {
    snmp_errno = SNMPERR_BAD_SESSION;
    return (0);
  }

  gettimeofday (&now,NULL);
  sp  = slp->session;
  isp = slp->internal;
  orp = NULL;
  API_DEBUG1 ("Checking session %s\n", sp->peername ? sp->peername : "<NULL>");

  for (rp = isp->requests; rp; rp = rp->next_request)
  {
    API_DEBUG4 ("Checking session request %d, expire at %u, Retry %d/%d\n",
                rp->request_id, rp->expire.tv_sec, rp->retries, sp->retries);
    if (freeme)
    {
      /* frees rp's after the for loop goes on to the next_request */
      free (freeme);
      freeme = NULL;
    }

    if (timercmp (&rp->expire, &now, <))
    {
      API_DEBUG0 ("API found an expired request.\n");

      /* this timer has expired
       */
      if (rp->retries >= sp->retries)
      {
        API_DEBUG0 ("No more chances, call the CallBack and delete this entry.\n");

        /* No more chances, delete this entry
         */
        if (sp->callback)
          (*sp->callback) (TIMED_OUT, sp, rp->pdu->reqid, rp->pdu,
                           sp->callback_magic);
        if (!orp)
             isp->requests     = rp->next_request;
        else orp->next_request = rp->next_request;

        snmp_free_pdu (rp->pdu);
        freeme = rp;
        continue;             /* don't update orp below */
      }
      else
      {
        BYTE   packet [PACKET_LENGTH];
        int    length = PACKET_LENGTH;
        struct timeval tv;

        /* retransmit this pdu */
        rp->retries++;
        API_DEBUG1 ("Retransmitting, %d try.\n", rp->retries);

        rp->timeout <<= 1;
        if (snmp_build (sp, rp->pdu, packet, &length) < 0)
           (*_printf) ("Error building packet\r\n");

        snmp_dump (packet, length, "sending", rp->pdu->address.sin_addr);
        gettimeofday (&tv, NULL);
        sock_fastwrite (isp->socket, (const char*)packet, length);

        rp->time    = tv;
        tv.tv_usec += rp->timeout;
        tv.tv_sec  += tv.tv_usec / 1000000L;
        tv.tv_usec %= 1000000L;
        rp->expire  = tv;
      }
    }
    orp = rp;
  }
  if (freeme)
     free (freeme);

  return (1);
}
