/* 
 * SNMP Network Packet Debugging
 *
 * Dumps the network packets to stdout.
 *
 */
/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * Author: Ryan Troll <ryan+@andrew.cmu.edu>
 *
 * Changes by G. Vanem 1998
 * Adapted to Waterloo tcp/ip
 **********************************************************************/

#include "snmp_lib.h"

static int _snmp_dump_packet = 0;

#ifdef USE_FORTIFY
static void ExitFortify (void)
{
  static int been_here = 0;

  if (been_here || _snmp_dump_packet <= 0)
     return;

  been_here = 1;
  Fortify_ListAllMemory();
  Fortify_OutputStatistics();
}
#endif

void snmp_dump_packet (int i)
{
  _snmp_dump_packet = i;

#ifdef USE_FORTIFY
  Fortify_EnterScope();
  atexit (ExitFortify);
#endif

}

static char *__inet_ntoa (struct in_addr adr)
{
  static char buf[30];
  char  *a = (char*)&adr;

  sprintf (buf, "%d.%d.%d.%d", a[0], a[1], a[2], a[3]);
  return (buf);
}

void snmp_dump (BYTE *buf, int buflen, char *how, struct in_addr who)
{
  int count, row;

  /* Return if there's nothing to do
   */
  if (_snmp_dump_packet == 0)
    return;

  printf ("%s %u bytes <-> %s:\n", how, buflen, __inet_ntoa(who));

  count = 0;
  while (count < buflen)
  {
    row = 0;
    for (; count + row < buflen && row < 16; row++)
        printf ("%02X ", buf[count + row]);

    while (row++ < 16)
       printf ("   ");
    printf ("  ");
    row = 0;
    for (; count + row < buflen && row < 16; row++)
    {
      int ch = buf [count+row];
      if (isprint(ch))
           printf ("%c", ch);
      else printf (".");
    }
    printf ("\n");
    count += row;
  }
  printf ("\n\n");
}

static char *TypeName (int type)
{
  switch (type)
  {
    case TYPE_OBJID:
         return ("objid");

    case TYPE_OCTETSTR:
         return ("oct-string");

    case TYPE_INTEGER:
         return ("integer");

    case TYPE_NETADDR:
         return ("net-addr");

    case TYPE_IPADDR:
         return ("ip-addr");

    case TYPE_COUNTER:
         return ("counter");

    case TYPE_GAUGE:
         return ("gauge");

    case TYPE_TIMETICKS:
         return ("timeticks");

    case TYPE_OPAQUE:
         return ("opaque");

    case TYPE_NULL:
         return ("null");

    case TYPE_OTHER:
    default:
         return ("?");
  }
}

void snmp_print_subtree (struct snmp_mib_tree *tree, int count)
{
  struct snmp_mib_tree *tp;

  printf ("%*s%s:\n", 2*count+1,"", tree->label);
  count++;
  for (tp = tree->child_list; tp; tp = tp->next_peer)
     printf ("%*s%s (%s)\n", 2*count+1, "", tp->label, TypeName(tp->type));

  for (tp = tree->child_list; tp; tp = tp->next_peer)
     snmp_print_subtree (tp, count);
}

