/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

Changes by G. Vanem 1998
Adapted to Waterloo tcp/ip

******************************************************************/

#include "snmp_lib.h"

/*
 *  local data/prototypes
 */
static void sprint_by_type (char *buf, struct variable_list *var,
                            struct enum_list *enums, int quiet);

static struct snmp_mib_tree *get_symbol (oid *objid, int objidlen,
                                         struct snmp_mib_tree *subtree,
                                         char *buf);

#ifdef __TURBOC__
#pragma warn -pro
#endif

static char *uptimeString (unsigned int timeticks, char *buf)
{
  int seconds, minutes, hours, days;

  timeticks /= 100;
  days = timeticks / (60 * 60 * 24);
  timeticks %= (60 * 60 * 24);

  hours = timeticks / (60 * 60);
  timeticks %= (60 * 60);

  minutes = timeticks / 60;
  seconds = timeticks % 60;

  switch (days)
  {
    case 0:  sprintf (buf, "%d:%02d:%02d", hours, minutes,seconds);
             break;
    case 1:  sprintf (buf, "%d day, %d:%02d:%02d", days,hours,minutes,seconds);
             break;
    default: sprintf (buf, "%d days, %d:%02d:%02d",days,hours,minutes,seconds);
  }
  return (buf);
}

static void sprint_hexstring (char *buf, BYTE *cp, int len)
{
  for (; len >= 16; len -= 16)
  {
    buf += sprintf (buf, "%02X %02X %02X %02X %02X %02X %02X %02X ",
                    cp[0], cp[1], cp[2], cp[3], cp[4], cp[5], cp[6], cp[7]);
    cp  += 8;
    buf += sprintf (buf, "%02X %02X %02X %02X %02X %02X %02X %02X\n",
                    cp[0], cp[1], cp[2], cp[3], cp[4], cp[5], cp[6], cp[7]);
    cp  += 8;
  }
  for (; len > 0; len--)
    buf += sprintf (buf, "%02X ", *cp++);
  *buf = '\0';
}

static void sprint_asciistring (char *buf, BYTE *cp, int len)
{
  int x;

  for (x = 0; x < len; x++)
  {
    if (isprint (*cp))
    {
      *buf++ = *cp++;
    }
    else
    {
      *buf++ = '.';
      cp++;
    }
    if ((x % 48) == 47)
      *buf++ = '\n';
  }
  *buf = '\0';
}

static void sprint_octet_string (char *buf, struct variable_list *var,
                                 struct enum_list *foo, int quiet)
{
  int    hex, x;
  BYTE *cp;

  if (var->type != ASN_OCTET_STR)
  {
    buf += sprintf (buf, "Wrong Type (should be OCTET STRING): ");
    sprint_by_type (buf,var,NULL,quiet);
    return;
  }
  hex = 0;
  for (cp = var->val.string, x = 0; x < var->val_len; x++, cp++)
  {
    if (!(isprint(*cp) || isspace(*cp)))
    {
      hex = 1;
      break;
    }
  }

  /* Now only check the last char if it's not a NULL.
   * Some things include the trailing NULL in the supplied length,
   * while others don't.
   *
   * If len>0 we have to check the last char
   */
  if (!hex && var->val_len && *cp &&
      !(isprint(*cp) || isspace(*cp)))
    hex = 1;

#if 0
  if (var->val_len <= 4)
     hex = 1;                 /* not likely to be ascii */
#endif

  if (hex)
  {
    if (!quiet)
       buf += sprintf (buf, "OCTET STRING-   (hex):\t");

    sprint_hexstring (buf, var->val.string, var->val_len);
  }
  else
  {
    if (!quiet)
      buf += sprintf (buf, "OCTET STRING- (ascii):\t");

    sprint_asciistring (buf, var->val.string, var->val_len);
  }
  ARGSUSED (foo);
}

static void sprint_opaque (char *buf, struct variable_list *var,
                           struct enum_list *foo, int quiet)
{
  if (var->type != SMI_OPAQUE)
  {
    buf += sprintf (buf, "Wrong Type (should be Opaque): ");
    sprint_by_type (buf,var,NULL,quiet);
    return;
  }
  if (!quiet)
    buf += sprintf (buf, "OPAQUE -   (hex):\t");

  sprint_hexstring (buf, var->val.string, var->val_len);
  ARGSUSED (foo);
}

static void sprint_object_identifier (char *buf, struct variable_list *var,
                                      struct enum_list *foo, int quiet)
{
  if (var->type != SMI_OBJID)
  {
    buf += sprintf (buf, "Wrong Type (should be OBJECT IDENTIFIER): ");
    sprint_by_type (buf, var, NULL, quiet);
    return;
  }
  if (!quiet)
     buf += sprintf (buf, "OBJECT IDENTIFIER:\t");

  sprint_objid (buf, (oid *)var->val.objid, var->val_len / sizeof(oid));
  ARGSUSED (foo);
}

static void sprint_timeticks (char *buf, struct variable_list *var,
                              struct enum_list *foo, int quiet)
{
  char timebuf[32];

  if (var->type != SMI_TIMETICKS)
  {
    buf += sprintf (buf, "Wrong Type (should be Timeticks): ");
    sprint_by_type (buf, var, NULL, quiet);
    return;
  }
  if (!quiet)
     buf += sprintf (buf, "Timeticks: ");

  sprintf (buf, "(%u) %s", *var->val.integer,
                           uptimeString(*var->val.integer,timebuf));
  ARGSUSED (foo);
}

static void sprint_integer (char *buf, struct variable_list *var,
                            struct enum_list *enums, int quiet)
{
  char *enum_string = NULL;

  if (var->type != SMI_INTEGER)
  {
    buf += sprintf (buf, "Wrong Type (should be INTEGER): ");
    sprint_by_type (buf, var, NULL, quiet);
    return;
  }
  for (; enums; enums = enums->next)
    if (enums->value == *var->val.integer)
    {
      enum_string = enums->label;
      break;
    }

  if (!quiet)
     buf += sprintf (buf, "INTEGER: ");

  if (!enum_string)
       sprintf (buf, "%u",     *var->val.integer);
  else sprintf (buf, "%s(%u)", enum_string, *var->val.integer);
}

static void sprint_gauge (char *buf, struct variable_list *var,
                          struct enum_list *foo, int quiet)
{
  if (var->type != SMI_GAUGE32)
  {
    buf += sprintf (buf, "Wrong Type (should be Gauge): ");
    sprint_by_type (buf, var, NULL, quiet);
    return;
  }
  if (!quiet)
     buf += sprintf (buf, "Gauge: ");

  sprintf (buf, "%u", *var->val.integer);
  ARGSUSED (foo);
}

static void sprint_counter (char *buf, struct variable_list *var,
                            struct enum_list *foo, int quiet)
{
  if (var->type != SMI_COUNTER32)
  {
    buf += sprintf (buf, "Wrong Type (should be Counter): ");
    sprint_by_type (buf, var, NULL, quiet);
    return;
  }
  if (!quiet)
     buf += sprintf (buf, "Counter: ");

  sprintf (buf, "%u", *var->val.integer);
  ARGSUSED (foo);
}

static void sprint_networkaddress (char *buf, struct variable_list *var,
                                   void *foo, int quiet)
{
  int   x, len;
  BYTE *cp;

  if (!quiet)
     buf += sprintf (buf, "Network Address:\t");

  cp  = var->val.string;
  len = var->val_len;
  for (x = 0; x < len; x++)
  {
    buf += sprintf (buf, "%02X", *cp++);
    if (x < (len - 1))
       *buf++ = ':';
  }
  ARGSUSED (foo);
}

static void sprint_ipaddress (char *buf, struct variable_list *var,
                              struct enum_list *foo, int quiet)
{
  BYTE *ip;

  if (var->type != SMI_IPADDRESS)
  {
    buf += sprintf (buf, "Wrong Type (should be Ipaddress): ");
    sprint_by_type (buf, var, NULL, quiet);
    return;
  }
  ip = var->val.string;
  if (!quiet)
     buf += sprintf (buf, "IPAddress:\t");

  sprintf (buf, "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);
  ARGSUSED (foo);
}


static void sprint_null (char *buf, struct variable_list *var,
                         struct enum_list *foo, int quiet)
{
  if (var->type != SMI_NULLOBJ)
  {
    buf += sprintf (buf, "Wrong Type (should be NULL): ");
    sprint_by_type (buf, var, NULL, quiet);
    return;
  }
  sprintf (buf, "NULL");
  ARGSUSED (foo);
}

static void sprint_unknowntype (char *buf, struct variable_list *var,
                                void *foo, int quiet)
{
/*sprintf(buf, "Variable has bad type"); */
  sprint_by_type (buf, var, foo, quiet);
}

static void sprint_badtype (char *buf)
{
  sprintf (buf, "Variable has bad type");
}

static void sprint_by_type (char *buf, struct variable_list *var,
                            struct enum_list *enums, int quiet)
{
  switch (var->type)
  {
    case SMI_INTEGER:
         sprint_integer (buf, var, enums, quiet);
         break;
    case SMI_STRING:
         sprint_octet_string (buf, var, enums, quiet);
         break;
    case SMI_OPAQUE:
         sprint_opaque (buf, var, enums, quiet);
         break;
    case SMI_OBJID:
         sprint_object_identifier (buf, var, enums, quiet);
         break;
    case SMI_TIMETICKS:
         sprint_timeticks (buf, var, enums, quiet);
         break;
    case SMI_GAUGE32:
         sprint_gauge (buf, var, enums, quiet);
         break;
    case SMI_COUNTER32:
         sprint_counter (buf, var, enums, quiet);
         break;
    case SMI_IPADDRESS:
         sprint_ipaddress (buf, var, enums, quiet);
         break;
    case SMI_NULLOBJ:
         sprint_null (buf, var, enums, quiet);
         break;
    default:
         sprint_badtype (buf);
         break;
  }
}



oid  RFC1066_MIB[]      = { 1, 3, 6, 1, 2, 1 };
char RFC1066_MIB_text[] = ".iso.org.dod.internet.mgmt.mib-2";

struct snmp_mib_tree *Mib;

static void set_functions (struct snmp_mib_tree *subtree)
{
  for (; subtree; subtree = subtree->next_peer)
  {
    switch (subtree->type)
    {
      case TYPE_OBJID:
           subtree->printer = sprint_object_identifier;
           break;
      case TYPE_OCTETSTR:
           subtree->printer = sprint_octet_string;
           break;
      case TYPE_INTEGER:
           subtree->printer = sprint_integer;
           break;
      case TYPE_NETADDR:
           subtree->printer = sprint_networkaddress;
           break;
      case TYPE_IPADDR:
           subtree->printer = sprint_ipaddress;
           break;
      case TYPE_COUNTER:
           subtree->printer = sprint_counter;
           break;
      case TYPE_GAUGE:
           subtree->printer = sprint_gauge;
           break;
      case TYPE_TIMETICKS:
           subtree->printer = sprint_timeticks;
           break;
      case TYPE_OPAQUE:
           subtree->printer = sprint_opaque;
           break;
      case TYPE_NULL:
           subtree->printer = sprint_null;
           break;
      case TYPE_OTHER:
      default:
           subtree->printer = sprint_unknowntype;
           break;
    }
    set_functions (subtree->child_list);
  }
}


/*
 * Config-file hook for WatTCP
 */
static void (*prev_hook)(const char *, const char *);

static int load_mib (const char *path, int MibType)
{
  if (Mib)
     return (1);

  if (MibType == OLD_CMU_SNMP_MIB)
     Mib = read_mib (path);

  else if (MibType == SNMPV2_MIB)
     Mib = read_mib_v2 (path);

  if (!Mib)
     return (0);

  set_functions (Mib);
  return (1);
}

static void get_mib1_file (const char *value)
{
  if (!load_mib(value, OLD_CMU_SNMP_MIB))
     fprintf (stderr, "SNMP: Failed to load `%s'\n", value);
}

static void get_mib2_file (const char *value)
{
  if (!load_mib(value, SNMPV2_MIB))
     fprintf (stderr, "SNMP: Failed to load `%s'\n", value);
}

static void config_hook (const char *name, const char *value)
{
  static struct config_table snmp_cfg[] = {
                { "MIBFILE",  ARG_FUNC, (void*)get_mib1_file },
                { "MIB2FILE", ARG_FUNC, (void*)get_mib2_file },
                { NULL }
              };

  if (!parse_config_table(&snmp_cfg[0], "SNMP.", name, value) && prev_hook)
     (*prev_hook) (name, value);
}

void init_mib (void)
{
  Mib = NULL;

  if (my_ip_addr)    /* not completely fool-proof! */
  {
    printf ("init_mib() must be called before sock_init()\r\n");
    exit (-1);
  }
  prev_hook = usr_init;
  usr_init  = config_hook;  /* Hook our config handler */
}


static struct snmp_mib_tree *find_rfc1066_mib (struct snmp_mib_tree *root)
{
  oid   *op = RFC1066_MIB;
  int    len;
  struct snmp_mib_tree *tp;

  for (len = sizeof (RFC1066_MIB) / sizeof (oid); len; len--, op++)
  {
    for (tp = root; tp; tp = tp->next_peer)
    {
      if (tp->subid == *op)
      {
        root = tp->child_list;
        break;
      }
    }
    if (!tp)
      return (NULL);
  }
  return (root);
}

static int lc_cmp (char *s1, char *s2)
{
  char c1, c2;

  while (*s1 && *s2)
  {
    if (isupper (*s1))
         c1 = tolower (*s1);
    else c1 = *s1;
    if (isupper (*s2))
         c2 = tolower (*s2);
    else c2 = *s2;
    if (c1 != c2)
       return ((c1 - c2) > 0 ? 1 : -1);
    s1++;
    s2++;
  }

  if (*s1)
     return (-1);
  if (*s2)
    return (1);
  return (0);
}

static int parse_subtree (struct snmp_mib_tree *subtree,
                          char *input,
                          oid *output,
                          int *out_len)  /* number of subid's */
{
  char   buf[128], *to = buf;
  DWORD  subid = 0;
  struct snmp_mib_tree *tp;

  /*
   * No empty strings.  Can happen if there is a trailing '.' or two '.'s
   * in a row, i.e. "..".
   */
  if (*input == '\0' || *input == '.')
     return (0);

  if (isdigit(*input))
  {
    /*
     * Read the number, then try to find it in the subtree.
     */
    while (isdigit (*input))
    {
      subid *= 10;
      subid += *input++ - '0';
    }
    for (tp = subtree; tp; tp = tp->next_peer)
    {
      if (tp->subid == subid)
        goto found;
    }
    tp = NULL;
  }
  else
  {
    /*
     * Read the name into a buffer.
     */
    while (*input != '\0' && *input != '.')
       *to++ = *input++;

    *to = '\0';

    /*
     * Find the name in the subtree;
     */
    for (tp = subtree; tp; tp = tp->next_peer)
    {
      if (lc_cmp (tp->label, buf) == 0)
      {
        subid = tp->subid;
        goto found;
      }
    }

    /*
     * If we didn't find the entry, punt...
     */
    if (!tp)
    {
      fprintf (stderr, "sub-identifier not found: %s\n", buf);
      return (0);
    }
  }

found:
  if (subid > (DWORD) MAX_SUBID)
  {
    fprintf (stderr, "sub-identifier too large: %s\n", buf);
    return (0);
  }

  if ((*out_len)-- <= 0)
  {
    fprintf (stderr, "object identifier too long\n");
    return (0);
  }
  *output++ = subid;

  if (*input != '.')
     return (1);

  if (tp)
       *out_len = parse_subtree (tp->child_list, ++input, output, out_len);
  else *out_len = parse_subtree (NULL, ++input, output, out_len);

  if (*out_len == 0)
    return (0);
  return (++*out_len);
}

int read_objid (char *input,
                oid  *output,
                int  *out_len)  /* number of subid's in "output" */
{
  struct snmp_mib_tree *root = Mib;
  oid   *op = output;
  int    i;

  if (*input == '.')
    input++;
  else
  {
    root = find_rfc1066_mib (root);
    for (i = 0; i < sizeof (RFC1066_MIB) / sizeof (oid); i++)
    {
      if ((*out_len)-- > 0)
        *output++ = RFC1066_MIB[i];
      else
      {
        fprintf (stderr, "object identifier too long\n");
        return (0);
      }
    }
  }

  if (!root)
  {
    oid *tmp;

    if (!mib_TxtToOid (input, &tmp, out_len))
    {
      fprintf (stderr, "Mib not initialized.  Exiting.\n");
      return (0);
    }
    memcpy (output, tmp, (*out_len * sizeof(oid)));
    free (tmp);
    return (1);
  }
  if ((*out_len = parse_subtree (root, input, output, out_len)) == 0)
     return (0);

  *out_len += output - op;
  return (1);
}

void print_objid (oid *objid,
                  int objidlen)  /* number of subidentifiers */
{
  char   buf[256];
  struct snmp_mib_tree *subtree = Mib;

  *buf = '.';                        /* this is a fully qualified name */
  get_symbol (objid, objidlen, subtree, buf+1);
  printf ("%s\n", buf);
}

void sprint_objid (char *buf,
                   oid  *objid,
                   int   objidlen)  /* number of subidentifiers */
{
  struct snmp_mib_tree *subtree = Mib;

  *buf = '.';                        /* this is a fully qualified name */
  get_symbol (objid, objidlen, subtree, buf+1);
}


void print_variable (oid *objid, int objidlen, struct variable_list *variable)
{
  char   buf[1024], *cp;
  struct snmp_mib_tree *subtree = Mib;

  *buf    = '.';              /* this is a fully qualified name */
  subtree = get_symbol (objid, objidlen, subtree, buf + 1);
  cp      = buf;

  if ((strlen(buf) >= strlen(RFC1066_MIB_text)) &&
      !memcmp(buf,RFC1066_MIB_text,strlen(RFC1066_MIB_text)))
     cp += sizeof (RFC1066_MIB_text);

  printf ("Name: %s -> ", cp);
  *buf = '\0';

  if (subtree)
  {
    if (subtree->printer)
         (*subtree->printer) (buf, variable, subtree->enums, 0);
    else sprint_by_type (buf, variable, subtree->enums, 0);
  }
  printf ("%s\n", buf);
}


void sprint_variable (char *buf, oid *objid, int objidlen,
                      struct variable_list *variable)
{
  char   tempbuf[512], *cp;
  struct snmp_mib_tree *subtree = Mib;

  *tempbuf = '.';                /* this is a fully qualified name */
  subtree  = get_symbol (objid, objidlen, subtree, tempbuf + 1);
  cp = tempbuf;
  if ((strlen(buf) >= strlen(RFC1066_MIB_text)) &&
      !memcmp(buf,RFC1066_MIB_text,strlen(RFC1066_MIB_text)))
    cp += sizeof (RFC1066_MIB_text);

  buf += sprintf (buf, "Name: %s -> ", cp);

  if (subtree)
  {
    if (subtree->printer)
         (*subtree->printer) (buf, variable, subtree->enums, 0);
    else sprint_by_type (buf, variable, subtree->enums, 0);
  }
  strcat (buf, "\n");
}

void sprint_value (char *buf, oid *objid, int objidlen,
                   struct variable_list *variable)
{
  char   tempbuf[512];
  struct snmp_mib_tree *subtree = Mib;

  subtree = get_symbol (objid, objidlen, subtree, tempbuf);

  if (subtree)
  {
    if (subtree->printer)
         (*subtree->printer) (buf, variable, subtree->enums, 0);
    else sprint_by_type (buf, variable, subtree->enums, 0);
  }
  else
  {
    buf[0] ='?';
    buf[1] = 0;
  }
}

void print_value (oid *objid, int objidlen, struct variable_list *variable)
{
  char   tempbuf[512];
  struct snmp_mib_tree *subtree = Mib;

  subtree = get_symbol (objid, objidlen, subtree, tempbuf);

  if (subtree)
  {
    if (subtree->printer)
         (*subtree->printer) (tempbuf, variable, subtree->enums, 0);
    else sprint_by_type (tempbuf, variable, subtree->enums, 0);
    printf ("%s\n", tempbuf);
  }
  else
    putchar ('?');
}

static struct snmp_mib_tree *get_symbol (oid *objid, int objidlen,
                                         struct snmp_mib_tree *subtree,
                                         char *buf)
{
  struct snmp_mib_tree *return_tree = NULL;

  for (; subtree; subtree = subtree->next_peer)
  {
    if (*objid == subtree->subid)
    {
      strcpy (buf, subtree->label);
      if (objidlen > 1)
      {
        while (*buf)
           buf++;
        *buf++ = '.';
        *buf   = '\0';
        return_tree = get_symbol (objid+1, objidlen-1,
                                  subtree->child_list, buf);
      }
      if (return_tree)
         return (return_tree);
      return (subtree);
    }
  }

  /* subtree not found
   */
  while (objidlen--)
  {                      /* output rest of name, uninterpreted */
    sprintf (buf, "%lu.", *objid++);
    while (*buf)
      buf++;
  }
  *(buf-1) = '\0';       /* remove trailing dot */
  return (NULL);
}


void print_variable_list (struct variable_list *V)
{
  print_variable (V->name, V->name_length, V);
}


void print_variable_list_value (struct variable_list *variable)
{
  char   buf[512];
  struct snmp_mib_tree *subtree = Mib;

  *buf    = '.';        /* this is a fully qualified name */
  subtree = get_symbol (variable->name, variable->name_length, subtree, buf+1);
  *buf    = '\0';

  if (subtree)
  {
    if (subtree->printer)
         (*subtree->printer) (buf, variable, subtree->enums, 1);
    else sprint_by_type (buf, variable, subtree->enums, 1);
  }
  printf ("%s", buf);
}

void print_type (struct variable_list *var)
{
  switch (var->type)
  {
    case SMI_INTEGER:
         printf ("Integer");
         break;
    case SMI_STRING:
         printf ("Octet String");
         break;
    case SMI_OPAQUE:
         printf ("Opaque");
         break;
    case SMI_OBJID:
         printf ("Object Identifier");
         break;
    case SMI_TIMETICKS:
         printf ("Timeticks");
         break;
    case SMI_GAUGE32:
         printf ("Gauge");
         break;
    case SMI_COUNTER32:
         printf ("Counter");
         break;
    case SMI_IPADDRESS:
         printf ("IP Address");
         break;
    case SMI_NULLOBJ:
         printf ("NULL");
         break;
    default:
         printf ("Unknown type %d\n", var->type);
         break;
  }
}

void print_oid_nums (oid * O, int len)
{
  int x;

  for (x = 0; x < len; x++)
    printf (".%lu", O[x]);
}

void free_mib (void)
{
  struct snmp_mib_tree *ptr, *next;

  for (ptr = Mib; ptr; ptr = next)
  {
    next = ptr->next_peer;
    tree_Free (ptr);
  }
}
