/**********************************************************************
 *
 *           Copyright 1998 by Carnegie Mellon University
 * 
 *                       All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * Author: Ryan Troll <ryan+@andrew.cmu.edu>
 *
 * Changes by G. Vanem 1998
 * Adapted to Waterloo tcp/ip
 **********************************************************************/

/*
 * SNMP PDU Encoding
 *
 * Complies with:
 *
 * RFC 1902: Structure of Management Information for SNMPv2
 *
 */
#include "snmp_lib.h"

#define ASN_PARSE_ERROR(x)  do {                         \
                              snmpInASNParseErrs_Add(1); \
                              return(x);                 \
                            } while (0)
/*
 * Create a PDU.
 */
struct snmp_pdu *snmp_pdu_create (int command)
{
  struct snmp_pdu *pdu = calloc (sizeof(*pdu),1);

  PDU_DEBUG0 ("Creating\n");
  if (!pdu)
  {
    snmp_set_api_error (SNMPERR_OS_ERR);
    return (NULL);
  }

  pdu->command                 = command;
  pdu->errstat                 = SNMP_ERR_NOERROR;
  pdu->address.sin_addr.s_addr = SNMP_DEFAULT_ADDRESS;
  PDU_DEBUG1 ("Created %p\n", pdu);
  return (pdu);
}

/*
 * Clone an existing PDU.
 */
struct snmp_pdu *snmp_pdu_clone (struct snmp_pdu *src)
{
  struct snmp_pdu *dest = malloc (sizeof(*dest));

  if (!dest)
  {
    snmp_set_api_error (SNMPERR_OS_ERR);
    return (NULL);
  }
  memcpy (dest,src,sizeof(*src));
  return (dest);
}

/* 
 * If there was an error in the input pdu, creates a clone of the pdu
 * that includes all the variables except the one marked by the errindex.
 * The command is set to the input command and the reqid, errstat, and
 * errindex are set to default values.
 * If the error status didn't indicate an error, the error index didn't
 * indicate a variable, the pdu wasn't a get response message, or there
 * would be no remaining variables, this function will return NULL.
 * If everything was successful, a pointer to the fixed cloned pdu will
 * be returned.
 */
struct snmp_pdu *snmp_pdu_fix (struct snmp_pdu *pdu, int command)
{
  return (snmp_fix_pdu (pdu, command));
}

struct snmp_pdu *snmp_fix_pdu (struct snmp_pdu *pdu, int command)
{
  struct variable_list *var, *newvar;
  struct snmp_pdu      *newpdu;
  int    index;
  int    copied = 0;

  PDU_DEBUG2 ("PDU %p:  Fixing.  Err index is %d\n",
              pdu, (unsigned int) pdu->errindex);

  if (pdu->command != SNMP_PDU_RESPONSE ||
      pdu->errstat == SNMP_ERR_NOERROR  ||
      pdu->errindex <= 0)
  {
    snmp_set_api_error (SNMPERR_UNABLE_TO_FIX);
    return (NULL);
  }

  /* clone the pdu
   */
  newpdu = snmp_pdu_clone (pdu);
  if (!newpdu)
     return (NULL);

  newpdu->variables = 0;
  newpdu->command   = command;
  newpdu->reqid     = SNMP_DEFAULT_REQID;
  newpdu->errstat   = SNMP_ERR_NOERROR;
  newpdu->errindex  = 0;         /* XXXXX Is this right? */

  /* Loop through the variables, removing whatever isn't necessary
   */
  var   = pdu->variables;
  index = 1;

  /* skip first variable if necessary
   */
  if (pdu->errindex == index)
  {
    var = var->next_variable;
    index++;
  }

  if (var)
  {
    /* VAR is the first uncopied variable
     *
     * Clone this variable
     */
    newpdu->variables = snmp_var_clone (var);
    if (!newpdu->variables)
    {
      snmp_pdu_free (newpdu);
      return (NULL);
    }
    copied++;
    newvar = newpdu->variables;

    /* VAR has been copied to NEWVAR.
     */
    while (var->next_variable)
    {
      /* Skip the item that was bad */
      if (++index == pdu->errindex)
      {
        var = var->next_variable;
        continue;
      }

      /* Copy this var
       */
      newvar->next_variable = snmp_var_clone (var->next_variable);
      if (!newvar->next_variable)
      {
        snmp_pdu_free (newpdu);
        return (NULL);
      }

      /* Move to the next one
       */
      newvar = newvar->next_variable;
      var    = var->next_variable;
      copied++;
    }
    newvar->next_variable = NULL;
  }

  /* If we didn't copy anything, free the new pdu.
   */
  if (index < pdu->errindex || copied == 0)
  {
    snmp_free_pdu (newpdu);
    snmp_set_api_error (SNMPERR_UNABLE_TO_FIX);
    return (NULL);
  }
  PDU_DEBUG2 ("PDU %p:  Fixed PDU is %p\n", pdu,newpdu);
  return (newpdu);
}


void snmp_pdu_free (struct snmp_pdu *pdu)
{
  snmp_free_pdu (pdu);
}

/* 
 * Frees the pdu and any malloc'd data associated with it.
 */
void snmp_free_pdu (struct snmp_pdu *pdu)
{
  struct variable_list *vp, *ovp;

  vp = pdu->variables;
  while (vp)
  {
    ovp = vp;
    vp = vp->next_variable;
    snmp_var_free (ovp);
  }
  if (pdu->enterprise)
    free (pdu->enterprise);
  free (pdu);
}


/* Encode this PDU into DestBuf.
 *
 * Returns a pointer to the next byte in the buffer (where the Variable
 * Bindings belong.)
 */

/* 
 * RFC 1902: Structure of Management Information for SNMPv2
 *
 *   PDU ::= 
 *    SEQUENCE {
 *      request-id   INTEGER32
 *      error-status INTEGER
 *      error-index  INTEGER
 *      Variable Bindings
 *    }
 *
 * BulkPDU ::=
 *    SEQUENCE {
 *      request-id      INTEGER32
 *      non-repeaters   INTEGER
 *      max-repetitions INTEGER
 *      Variable Bindings
 *    }
 */

/* 
 * RFC 1157: A Simple Network Management Protocol (SNMP)
 *
 *   PDU ::= 
 *    SEQUENCE {
 *      request-id   INTEGER
 *      error-status INTEGER
 *      error-index  INTEGER
 *      Variable Bindings
 *    }
 *
 *   TrapPDU ::=
 *    SEQUENCE {
 *      enterprise    NetworkAddress
 *      generic-trap  INTEGER
 *      specific-trap INTEGER
 *      time-stamp    TIMETICKS
 *      Variable Bindings
 *    }
 */

BYTE *snmp_pdu_encode (BYTE *DestBuf, int *DestBufLen,
                       struct snmp_pdu *PDU)
{
  BYTE *bufp;

  PDU_ENCODE_DEBUG1 ("Encoding %d\n", PDU->command);

  /* ASN.1 Header
   */
  switch (PDU->command)
  {
    case TRP_REQ_MSG:

         /* SNMPv1 Trap */

         /* enterprise */
         bufp = asn_build_objid (DestBuf, DestBufLen,
                                 (BYTE) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID),
                                 (oid*) PDU->enterprise, PDU->enterprise_length);
         if (!bufp)
           return (NULL);

         /* agent-addr */
         bufp = asn_build_string (bufp, DestBufLen,
                                  (BYTE) (SMI_IPADDRESS | ASN_PRIMITIVE),
                                  (BYTE*) &PDU->agent_addr.sin_addr.s_addr,
                                  sizeof(PDU->agent_addr.sin_addr.s_addr));
         if (!bufp)
           return (NULL);

         /* generic trap */
         bufp = asn_build_int (bufp, DestBufLen,
                               (BYTE) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                               (int*) &PDU->trap_type, sizeof(PDU->trap_type));
         if (!bufp)
           return (NULL);

         /* specific trap */
         bufp = asn_build_int (bufp, DestBufLen,
                               (BYTE) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                               (int*) &PDU->specific_type, sizeof(PDU->specific_type));
         if (!bufp)
           return (NULL);

         /* timestamp */
         bufp = asn_build_unsigned_int (bufp, DestBufLen,
                                        (BYTE) (SMI_TIMETICKS | ASN_PRIMITIVE),
                                        &PDU->time, sizeof(PDU->time));
         if (!bufp)
           return (NULL);
         break;

    case SNMP_PDU_GETBULK:

         /* SNMPv2 Bulk Request */

         /* request id */
         bufp = asn_build_int (DestBuf, DestBufLen,
                               (BYTE) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                               &PDU->reqid, sizeof(PDU->reqid));
         if (!bufp)
           return (NULL);

         /* non-repeaters */
         bufp = asn_build_int (bufp, DestBufLen,
                               (BYTE) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                               &PDU->non_repeaters, sizeof(PDU->non_repeaters));
         if (!bufp)
           return (NULL);

         /* max-repetitions */
         bufp = asn_build_int (bufp, DestBufLen,
                               (BYTE) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                               &PDU->max_repetitions, sizeof(PDU->max_repetitions));
         if (!bufp)
           return (NULL);
         break;

    default:

         /* Normal PDU Encoding */

         /* request id */
         PDU_ENCODE_DEBUG2 ("Request ID %d (0x%p)\n", PDU->reqid, DestBuf);
         bufp = asn_build_int (DestBuf, DestBufLen,
                               (BYTE) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                               &PDU->reqid, sizeof(PDU->reqid));
         if (!bufp)
           return (NULL);

         /* error status */
         PDU_ENCODE_DEBUG2 ("Error Status %d (0x%p)\n", PDU->errstat, bufp);
         bufp = asn_build_int (bufp, DestBufLen,
                               (BYTE) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                               &PDU->errstat, sizeof(PDU->errstat));
         if (!bufp)
           return (NULL);

         /* error index */
         PDU_ENCODE_DEBUG2 ("Error index %d (0x%p)\n", PDU->errindex, bufp);
         bufp = asn_build_int (bufp, DestBufLen,
                               (BYTE) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                               &PDU->errindex, sizeof(PDU->errindex));
         if (!bufp)
           return (NULL);
         break;
  }
  return (bufp);
}


/* Decodes PDU from Packet into PDU.
 *
 * Returns a pointer to the next byte of the packet, which is where the
 * Variable Bindings start.
 */
BYTE *snmp_pdu_decode (BYTE            *Packet,   /* data */
                       int             *Length,   /* &length */
                       struct snmp_pdu *PDU)      /* pdu */
{
  BYTE *bufp;
  BYTE  PDUType;
  int   four;
  BYTE  ASNType;
  oid   objid[MAX_NAME_LEN];

  bufp = asn_parse_header (Packet, Length, &PDUType);
  if (!bufp)
     ASN_PARSE_ERROR (NULL);

  PDU_DECODE_DEBUG1 ("Type %d\n", PDUType);
  PDU->command = PDUType;

  switch (PDUType)
  {
    case TRP_REQ_MSG:
         /* SNMPv1 Trap Message */

         /* enterprise */
         PDU->enterprise_length = MAX_NAME_LEN;
         bufp = asn_parse_objid (bufp, Length, &ASNType, objid,
                                 &PDU->enterprise_length);
         if (!bufp)
           ASN_PARSE_ERROR (NULL);

         PDU->enterprise = malloc (PDU->enterprise_length*sizeof(oid));
         if (!PDU->enterprise)
         {
           snmp_set_api_error (SNMPERR_OS_ERR);
           return (NULL);
         }

         memcpy (PDU->enterprise, objid, PDU->enterprise_length*sizeof(oid));

         /* Agent-addr */
         four = 4;
         bufp = asn_parse_string (bufp, Length,&ASNType,
                                  (BYTE*) &PDU->agent_addr.sin_addr.s_addr,
                                  &four);
         if (!bufp)
            ASN_PARSE_ERROR (NULL);

         /* Generic trap */
         bufp = asn_parse_int (bufp, Length, &ASNType,
                               (int*) &PDU->trap_type, sizeof(PDU->trap_type));
         if (!bufp)
            ASN_PARSE_ERROR (NULL);

         /* Specific Trap */
         bufp = asn_parse_int (bufp, Length, &ASNType,
                               (int*) &PDU->specific_type,
                               sizeof(PDU->specific_type));
         if (!bufp)
            ASN_PARSE_ERROR (NULL);

         /* Timestamp */
         bufp = asn_parse_unsigned_int (bufp, Length, &ASNType,
                                        &PDU->time, sizeof(PDU->time));
         if (!bufp)
            ASN_PARSE_ERROR (NULL);
         break;

    case SNMP_PDU_GETBULK:

         /* SNMPv2 Bulk Request */

         /* request id */
         bufp = asn_parse_int (bufp, Length, &ASNType,
                               &PDU->reqid, sizeof(PDU->reqid));
         if (!bufp)
            ASN_PARSE_ERROR (NULL);

         /* non-repeaters */
         bufp = asn_parse_int (bufp, Length, &ASNType,
                               &PDU->non_repeaters, sizeof(PDU->non_repeaters));
         if (!bufp)
            ASN_PARSE_ERROR (NULL);

         /* max-repetitions */
         bufp = asn_parse_int (bufp, Length, &ASNType,
                               &PDU->max_repetitions, sizeof(PDU->max_repetitions));
         if (!bufp)
            ASN_PARSE_ERROR (NULL);
         break;

    default:

         /* Normal PDU Encoding */

         /* request id */
         bufp = asn_parse_int (bufp, Length, &ASNType,
                               &PDU->reqid, sizeof (PDU->reqid));
         if (!bufp)
            ASN_PARSE_ERROR (NULL);

         PDU_DECODE_DEBUG1 ("Request ID %d\n", PDU->reqid);

         /* error status */
         bufp = asn_parse_int (bufp, Length,
                               &ASNType,
                               &PDU->errstat, sizeof (PDU->errstat));
         if (!bufp)
            ASN_PARSE_ERROR (NULL);

         PDU_DECODE_DEBUG1 ("Error Status %d\n", PDU->errstat);

         /* error index */
         bufp = asn_parse_int (bufp, Length,
                               &ASNType,
                               &PDU->errindex, sizeof (PDU->errindex));
         if (!bufp)
            ASN_PARSE_ERROR (NULL);

         PDU_DECODE_DEBUG1 ("Error Index %d\n", PDU->errindex);
         break;
  }
  return (bufp);
}


char *snmp_pdu_type (struct snmp_pdu *PDU)
{
  switch (PDU->command)
  {
    case SNMP_PDU_GET:
         return ("GET");
    case SNMP_PDU_GETNEXT:
         return ("GETNEXT");
    case SNMP_PDU_RESPONSE:
         return ("RESPONSE");
    case SNMP_PDU_SET:
         return ("SET");
    case SNMP_PDU_GETBULK:
         return ("GETBULK");
    case SNMP_PDU_INFORM:
         return ("INFORM");
    case SNMP_PDU_V2TRAP:
         return ("V2TRAP");
    case SNMP_PDU_REPORT:
         return ("REPORT");
    case TRP_REQ_MSG:
         return ("V1TRAP");
    default:
         return ("Unknown");
  }
}

/* 
 * Add a null variable with the requested name to the end of the list of
 * variables for this pdu.
 */
void snmp_add_null_var (struct snmp_pdu *pdu, oid *name, int name_length)
{
  struct variable_list *ptr;
  struct variable_list *vars = snmp_var_new (name, name_length);

  if (!vars)
  {
    perror ("snmp_add_null_var: malloc");
    return;
  }

  if (!pdu->variables)
     pdu->variables = vars;
  else
  {
    /* Insert at the end */
    for (ptr = pdu->variables;
         ptr->next_variable;
         ptr = ptr->next_variable)
        ;
    ptr->next_variable = vars;
  }
}
